/*
 * Copyright (c) 2008, Laurent Desnogues <laurent.desnogues@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the <organization> nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <assert.h>

/* max number of inputs (r0-r3) */
#define MAX_INS   4
/* max number of outputs (r4-r7) */
#define MAX_OUTS  4

static FILE *s_fin, *s_fout_c, *s_fout_asm;

static void cleanup(void)
{
  if (s_fin != NULL) {
    fclose(s_fin);
    s_fin = NULL;
  }
  if (s_fout_c != NULL) {
    fclose(s_fout_c);
    s_fout_c = NULL;
  }
  if (s_fout_asm != NULL) {
    fclose(s_fout_asm);
    s_fout_asm = NULL;
  }
}

static void error(const char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  fputc('\n', stderr);
  cleanup();
  exit(1);
}

/* remove EOL and trailing spaces */
static void clean_end_of_string(char *buf)
{
  int i;

  if (strlen(buf) != 0 && buf[strlen(buf) - 1] == '\n')
    buf[strlen(buf) - 1] = '\0';
  for (i = strlen(buf) - 1; i >= 0 && isspace(buf[i]); i--)
    buf[i] = '\0';
}

/* skip leading blanks */
static const char *skip_spaces(const char *buf)
{
  while (isspace(*buf))
    buf++;
  return buf;
}

/* buffer size for I/O */
#define BUF_SIZE  1024

/* input pattern structure */
typedef struct {
  const char *function_name;  /* label used for the function */
  const char *instruction;    /* assembly instruction */
  unsigned    nb_ins;         /* number of read registers */
  unsigned    nb_outs;        /* number of written registers */
  int         is_apsr_in;     /* is APSR part of the inputs */
  int         is_apsr_out;    /* should APSR be output */
  int         is_accumulate;  /* outputs need init */
} pattern_header_t;

typedef struct {
  uint32_t inputs[MAX_INS];
} pattern_t;

/* Copy to 'word' from 'in' until a space of end of string is seen.
 * Return:
 *  - NULL if no space/EOS was found
 *  - pointer to first non white char after word
 */
static const char *get_word(const char *in, char *word)
{
  if (isspace(*in) || *in == '\0')
    return NULL;
  /* copy word */
  do {
    *word++ = *in++;
  } while (!isspace(*in) && *in != '\0');
  *word = '\0';
  in = skip_spaces(in);
  return in;
}

static const char *get_optional_flags(const char *in, int *accumulate)
{
  if (*in == 'a') {
    *accumulate = 1;
    return skip_spaces(in + 1);
  }
  *accumulate = 0;
  return in;
}

static const char *get_decimal_flags(const char *in,
                                     unsigned *p_nb, int *apsr)
{
  char        buf[BUF_SIZE];
  const char *p_buf;
  const char *p_res;
  unsigned    nb;

  p_res = get_word(in, buf);
  if (p_res == NULL)
    return NULL;
  /* TODO check overflow */
  for (nb = 0, p_buf = buf; isdigit(*p_buf); p_buf++)
    nb = nb * 10 + (*p_buf - '0');
  if (*p_buf == '+') {
    *apsr = 1;
    p_buf++;
  } else
    *apsr = 0;
  if (*p_buf != '\0')
    return NULL;
  *p_nb = nb;
  return p_res;
}

static const char *get_hexa(const char *in, unsigned *p_nb)
{
  char        buf[BUF_SIZE];
  const char *p_buf;
  const char *p_res;
  unsigned    nb;

  p_res = get_word(in, buf);
  if (p_res == NULL)
    return NULL;
  /* TODO check overflow */
  for (nb = 0, p_buf = buf; isxdigit(*p_buf); p_buf++) {
    int c = tolower(*p_buf);

    nb = nb * 16 + c - (c >= 'a' ? 'a' - 10 : '0');
  }
  if (*p_buf != '\0')
    return NULL;
  *p_nb = nb;
  return p_res;
}

/* register usage
 *  r0-r3 input parameters
 *  r4-r7 outputs
 *  r8    pointer to result_buffer
 *  r9    r/w for APSR
 */

/* save/restore registers, enforcing 8 byte alignment */
/* NOTE this is done in a brutal dumb way */
static void callee_save_restore(FILE *fout_asm, int is_save,
                                unsigned nb_outs, int apsr)
{
#if MAX_OUTS > 4
#error MAX_OUTS too large
#endif
  assert(nb_outs < 4);
  fputs(is_save ? "\tpush\t" : "\tpop\t", fout_asm);
  if (apsr) {
    switch (nb_outs) {
    case 0:
      fprintf(fout_asm, "{r8,r9}\n");
      break;
    case 1: case 2:
      fprintf(fout_asm, "{r4-r5,r8,r9}\n");
      break;
    case 3: case 4:
      fprintf(fout_asm, "{r4-r7,r8,r9}\n");
      break;
    }
  } else {
    switch (nb_outs) {
    case 0: case 1:
      fprintf(fout_asm, "{r4,r8}\n");
      break;
    case 2: case 3:
      fprintf(fout_asm, "{r4-r6,r8}\n");
      break;
    case 4:
      fprintf(fout_asm, "{r4-r7,r8,r9}\n");
      break;
    }
  }
}

/* gas/armasm portability */
static const char  port_cmt[2] = { '@', ';' };
static const char *port_export[2] = { ".global", "EXPORT" };
static const char *port_label[2] = { ":", "" };
static const char *port_text[2] = { ".text", "AREA\t|.text|,CODE" };
static const char *port_word[2] = { ".word", "DCD" };
static const char *port_end[2] = { ".end", "END" };

static int generate_pattern(FILE *fin, FILE *fout_c, FILE *fout_asm,
                            const pattern_header_t *pattern_header,
                            int arm_syntax)
{
  char        in_buf[BUF_SIZE];
  const char *p_buf, *p_buf2;
  unsigned    nb_outs_with_apsr;
  unsigned    nb_ins_with_accumulate;
  unsigned    nb_patterns;
  pattern_t   pattern;
  int         i;

  // TODO input flag handling
  /* dump the asm function */
  fprintf(fout_asm, "%c Instruction '%s'\n",
          port_cmt[arm_syntax],
          pattern_header->instruction);
  fprintf(fout_asm, "\t%s\n", port_text[arm_syntax]);
  fprintf(fout_asm, "\t%s\t%s\n",
          port_export[arm_syntax], pattern_header->function_name);
  fprintf(fout_asm, "%s%s\n",
          pattern_header->function_name, port_label[arm_syntax]);
  // save registers, enforcing 8 byte alignment
  callee_save_restore(fout_asm, 1,
                      pattern_header->nb_outs, pattern_header->is_apsr_out);
  // copy extra inputs for accumulation if needed
  if (pattern_header->is_accumulate) {
    for (i = 0; i < pattern_header->nb_outs; i++)
      fprintf(fout_asm, "\tmov\tr%u, r%u\n",
              4 + i, pattern_header->nb_ins + i);
  }
  // dump instruction outs ins
  fprintf(fout_asm, "\t%s\t", pattern_header->instruction);
  for (i = 0; i < pattern_header->nb_outs; i++)
    fprintf(fout_asm, "%sr%d", i == 0 ? "" : ",", i + 4);
  if (pattern_header->nb_outs != 0)
    fputs(", ", fout_asm);
  for (i = 0; i < pattern_header->nb_ins; i++)
    fprintf(fout_asm, "%sr%d", i == 0 ? "" : ",", i);
  fputc('\n', fout_asm);
  // store outputs
  if (pattern_header->is_apsr_out != 0)
    fprintf(fout_asm, "\tmrs\tr9, cpsr\n");
  nb_outs_with_apsr = pattern_header->nb_outs + pattern_header->is_apsr_out;
  if (nb_outs_with_apsr != 0) {
    if (arm_syntax)
      // TODO fix this
      fprintf(fout_asm, "\tldr\tr8,[pc,#8]\n");
    else
      fprintf(fout_asm, "\tldr\tr8,1f\n");
    // TODO wrong for 0 output
    if (pattern_header->nb_outs == 0) {
      if (pattern_header->is_apsr_out != 0)
        fprintf(fout_asm, "\tstm\tr8,{r9}\n");
    } else if (pattern_header->nb_outs == 1)
      fprintf(fout_asm, "\tstm\tr8,{r4%s}\n",
              pattern_header->is_apsr_out ? ",r9" : "");
    else
      fprintf(fout_asm, "\tstm\tr8,{r4-r%u%s}\n", 3 + pattern_header->nb_outs,
              pattern_header->is_apsr_out ? ",r9" : "");
  }
  // restore registers
  callee_save_restore(fout_asm, 0,
                      pattern_header->nb_outs, pattern_header->is_apsr_out);
  // return
  fprintf(fout_asm, "\tbx\tlr\n");
  // if needed put address of output store buffer
  if (pattern_header->nb_outs != 0)
    fprintf(fout_asm, "1%s\n\t%s\tresult_buffer\n",
            port_label[arm_syntax], port_word[arm_syntax]);

  /* prototype to call the asm function */
  nb_ins_with_accumulate =
    pattern_header->nb_ins
    + (pattern_header->is_accumulate != 0 ? pattern_header->nb_outs : 0);
  fprintf(fout_c, "extern void %s(", pattern_header->function_name);
  for (i = 0; i < nb_ins_with_accumulate; i++)
    fprintf(fout_c, "%suint32_t r%d", i == 0 ? "": ", ", i);
  fprintf(fout_c, ");\n");

  /* dump calls to asm function */
  nb_patterns = 0;
  while (fgets(in_buf, BUF_SIZE, fin) != NULL) {
    clean_end_of_string(in_buf);
    p_buf = skip_spaces(in_buf);
    /* skip comments and empty lines */
    if (*p_buf == '#' || *p_buf == '\0')
      continue;
    /* check end */
    if (strcmp(p_buf, "}") == 0)
      break;
    nb_patterns++;
    /* retrieve inputs */
    for (i = 0; i < pattern_header->nb_ins; i++) {
      p_buf2 = get_hexa(p_buf, &pattern.inputs[i]);
      if (p_buf2 == NULL)
        error("Expected hex number: <%s>", p_buf);
      p_buf = p_buf2;
    }
    if (pattern_header->is_accumulate) {
      for (i = 0; i < pattern_header->nb_outs; i++) {
        p_buf2 = get_hexa(p_buf, &pattern.inputs[pattern_header->nb_ins + i]);
        if (p_buf2 == NULL)
          error("Expected hex number: <%s>", p_buf);
        p_buf = p_buf2;
      }
    }
    /* check we reached EOL */
    if (*p_buf != '\0' && *p_buf != '#')
      error("Garbage at end of line: <%s>", p_buf);
    /* dump call to function */
    fprintf(fout_c, "%s(", pattern_header->function_name);
    for (i = 0; i < nb_ins_with_accumulate; i++)
      fprintf(fout_c, "%s0x%08x", i == 0 ? "" : ", ", pattern.inputs[i]);
    if (pattern_header->is_accumulate) {
    }
    fputs(");\n", fout_c);
    /* dump outputs */
    fprintf(fout_c, "dump_res(\"%s", pattern_header->instruction);
    for (i = 0; i < nb_ins_with_accumulate; i++)
      fprintf(fout_c, " %08x", pattern.inputs[i]);
    fprintf(fout_c, " -> \", %u, %d);\n",
            pattern_header->nb_outs, pattern_header->is_apsr_out);
  }

  return nb_patterns;
}

static int generate(FILE *fin, FILE *fout_c, FILE *fout_asm, int arm_syntax)
{
  char              in_buf[BUF_SIZE];
  const char       *p_buf, *p_buf2;
  unsigned          nb_instrs;
  unsigned          nb_patterns;
  char              f_name[BUF_SIZE], i_name[BUF_SIZE];
  pattern_header_t  pattern_header;
  int               ret;

  if (arm_syntax) {
    fprintf(fout_asm, "\tEXTERN\tresult_buffer\n");
  }
  nb_instrs = 0;
  nb_patterns = 0;
  while (fgets(in_buf, BUF_SIZE, fin) != NULL) {
    clean_end_of_string(in_buf);
    p_buf = skip_spaces(in_buf);
    /* skip comments and empty lines */
    if (*p_buf == '#' || *p_buf == '\0')
      continue;
    nb_instrs++;
    /* retrieve function and instruction names */
    p_buf2 = get_word(p_buf, f_name);
    if (p_buf2 == NULL)
      error("Expected function name: <%s>", p_buf);
    p_buf = get_word(p_buf2, i_name);
    if (p_buf == NULL)
      error("Expected instruction name: <%s>", p_buf2);
    pattern_header.function_name = f_name;
    pattern_header.instruction = i_name;
    pattern_header.is_apsr_in = 0;
    pattern_header.is_apsr_out = 0;
    /* retrieve number of ins and outs */
    p_buf2 = get_decimal_flags(p_buf,
                               &pattern_header.nb_ins,
                               &pattern_header.is_apsr_in);
    if (p_buf2 == NULL)
      error("Expected number of inputs: <%s>", p_buf);
    p_buf = get_decimal_flags(p_buf2,
                              &pattern_header.nb_outs,
                              &pattern_header.is_apsr_out);
    if (p_buf == NULL)
      error("Expected number of outputs: <%s>", p_buf2);
    p_buf2 = get_optional_flags(p_buf, &pattern_header.is_accumulate);
    if (strcmp(p_buf2, "{") != 0)
      error("Expected '{': <%s>", p_buf2);
    /* some checks */
    if (pattern_header.nb_ins > MAX_INS)
      error("Too many inputs %u (max %u)", pattern_header.nb_ins, MAX_INS);
    if (pattern_header.nb_outs > MAX_OUTS)
      error("Too many outputs %u (max %u)", pattern_header.nb_outs, MAX_OUTS);
    if (pattern_header.is_accumulate
        && pattern_header.nb_ins + pattern_header.nb_outs > MAX_INS)
      error("Too many inputs with accumulation %u + %u (max %u)",
            pattern_header.nb_ins, pattern_header.nb_outs, MAX_INS);
    /* read each pattern and generate code */
    ret = generate_pattern(fin, fout_c, fout_asm, &pattern_header, arm_syntax);
    if (ret < 0)
      return 1;
    nb_patterns += ret;
  }
  fprintf(fout_asm, "\n\t%s\n", port_end[arm_syntax]);
  printf("Generated %u instructions and %u patterns.\n",
         nb_instrs, nb_patterns);
  return 0;
}

static void usage(const char *progname)
{
  fprintf(stderr, "Usage:  %s [-arm] input c_output asm_output\n", progname);
  exit(1);
}

int main(int argc, const char *argv[])
{
  const char *input_name, *output_name;
  char        output_c_name[PATH_MAX], output_asm_name[PATH_MAX];
  int         ret;
  int         arm_syntax;

  /* command line stuff */
  if (argc != 3
      && (argc == 4 && strcmp(argv[1], "-arm") != 0))
    usage(argv[0]);
  if (argc == 4) {
    arm_syntax = 1;
    argv++;
  } else
    arm_syntax = 0;
  input_name = argv[1];
  output_name = argv[2];

  /* open files */
  sprintf(output_c_name, "%s.c", output_name);
  sprintf(output_asm_name, "%s.S", output_name);
  s_fin = fopen(input_name, "r");
  if (s_fin == NULL) {
    fprintf(stderr, "Could not open input file <%s>.", input_name);
    return 1;
  }
  s_fout_c = fopen(output_c_name, "w");
  if (s_fout_c == NULL) {
    fprintf(stderr, "Could not open C output file <%s>.", output_c_name);
    return 1;
  }
  s_fout_asm = fopen(output_asm_name, "w");
  if (s_fout_asm == NULL) {
    fprintf(stderr, "Could not open asm output file <%s>.", output_asm_name);
    return 1;
  }

  /* generate */
  ret = generate(s_fin, s_fout_c, s_fout_asm, arm_syntax);

  cleanup();

  return ret;
}
